import time
import typing

# ROS2
import builtin_interfaces.msg
import rclpy
from rclpy.node import Node

from std_msgs.msg import Float64

# Ignition
from ignition.msgs.float_pb2 import Float
from ignition.msgs.model_pb2 import Model
from ignition.msgs.vector3d_pb2 import Vector3d
from ignition.msgs.wrench_pb2 import Wrench
from ignition.transport import Node as IgnNode
from ignition.transport import SubscribeOptions


# Joint state publishers (ROS2)
class JointStatePublishers(object):
    def __init__(self, ros2_node, model_name, joint_name):
        super().__init__()
        prefix = model_name + "/" + joint_name
        self.ros2_node = ros2_node
        self.model_name = model_name
        self.joint_name = joint_name
        self.position_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/position", 10)
        self.velocity_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/velocity", 10)
        self.force_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/force", 10)
        self.acceleration_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/acceleration", 10)

    def pub_position(self, stamp, value):
        msg = Float64()
        msg.data = value
        self.position_publisher.publish(msg)

    def pub_velocity(self, stamp, value):
        msg = Float64()
        msg.data = value
        self.velocity_publisher.publish(msg)

    def pub_force(self, stamp, value):
        msg = Float64()
        msg.data = value
        self.force_publisher.publish(msg)

    def pub_acceleration(self, stamp, value):
        msg = Float64()
        msg.data = value
        self.acceleration_publisher.publish(msg)


# PID publishers (ROS2)
class PIDPublishers(object):
    def __init__(self, ros2_node, model_name, joint_name):
        super().__init__()
        prefix = model_name + "/" + joint_name
        self.ros2_node = ros2_node
        self.model_name = model_name
        self.joint_name = joint_name
        self.act_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/pid/act", 10)
        self.tgt_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/pid/tgt", 10)
        self.cmd_publisher = self.ros2_node.create_publisher(
            Float64, prefix + "/pid/cmd", 10)

    def pub_act(self, value):
        msg = Float64()
        msg.data = value
        self.act_publisher.publish(msg)

    def pub_tgt(self, value):
        msg = Float64()
        msg.data = value
        self.tgt_publisher.publish(msg)

    def pub_cmd(self, value):
        msg = Float64()
        msg.data = value
        self.cmd_publisher.publish(msg)


# PID subscribers (Ignition)
class PIDSubscribers(object):
    def __init__(self, ros2_node, model_name, joint_name):
        super().__init__()

        self.ros2_node = ros2_node
        self.model_name = model_name
        self.joint_name = joint_name
        
        self.ign_node = IgnNode()
        self.last_pid_act = Float()
        self.last_pid_tgt = Float()
        self.last_pid_cmd = Float()

        self._sub_to_pid(model_name, joint_name)

    def _subscribe(self, topic, callback, msg_type_name, opts):
        if self.ign_node.subscribe(topic, callback, msg_type_name, opts):
            self.ros2_node.get_logger().info(
                "Subscribing to type [{}] on topic [{}]".format(
                    msg_type_name, topic))
        else:
            self.ros2_node.get_logger().error(
                "Error subscribing to topic [{}]".format(topic))

    def _sub_to_pid(self, model_name, joint_name):
        act_topic = "/model/{}/joint/{}/pid/act".format(
            model_name, joint_name)
        tgt_topic = "/model/{}/joint/{}/pid/tgt".format(
            model_name, joint_name)
        cmd_topic = "/model/{}/joint/{}/pid/cmd".format(
            model_name, joint_name)

        msg_type_name = Float.DESCRIPTOR.full_name
        opts = SubscribeOptions()

        # callbacks to capture the pid msgs
        def act_cb(msg: Float) -> None:
            self.last_pid_act = msg
        
        def tgt_cb(msg: Float) -> None:
            self.last_pid_tgt = msg

        def cmd_cb(msg: Float) -> None:
            self.last_pid_cmd = msg

        self._subscribe(act_topic, act_cb, msg_type_name, opts)
        self._subscribe(tgt_topic, tgt_cb, msg_type_name, opts)
        self._subscribe(cmd_topic, cmd_cb, msg_type_name, opts)


# LiftDrag subscribers (Ignition)
class LiftDragSubscribers(object):
    def __init__(self, ros2_node, model_name, link_name):
        super().__init__()

        self.ros2_node = ros2_node
        self.model_name = model_name
        self.link_name = link_name
        
        self.ign_node = IgnNode()
        self.last_lift = Vector3d()
        self.last_drag = Vector3d()
        self.last_force = Vector3d()
        self.last_torque = Vector3d()

        self._sub_to_lift(model_name, link_name)

    def _subscribe(self, topic, callback, msg_type_name, opts):
        if self.ign_node.subscribe(topic, callback, msg_type_name, opts):
            self.ros2_node.get_logger().info(
                "Subscribing to type [{}] on topic [{}]".format(
                    msg_type_name, topic))
        else:
            self.ros2_node.get_logger().error(
                "Error subscribing to topic [{}]".format(topic))

    def _sub_to_lift(self, model_name, link_name):
        act_topic = "/model/{}/link/{}/lift_drag/lift".format(
            model_name, link_name)

        msg_type_name = Vector3d.DESCRIPTOR.full_name
        opts = SubscribeOptions()

        # callbacks to capture the pid msgs
        def act_cb(msg: Float) -> None:
            self.last_pid_act = msg
        
        def tgt_cb(msg: Float) -> None:
            self.last_pid_tgt = msg

        def cmd_cb(msg: Float) -> None:
            self.last_pid_cmd = msg

        self._subscribe(act_topic, act_cb, msg_type_name, opts)
        self._subscribe(tgt_topic, tgt_cb, msg_type_name, opts)
        self._subscribe(cmd_topic, cmd_cb, msg_type_name, opts)


class ModelBridge(Node):

    def __init__(self):
        super().__init__('model_bridge')

        # world_name = "omnicopter_runway"
        # model_name = "omnicopter"

        # world_name = "alti_transition_runway"
        # model_name = "alti_transition_quad"

        world_name = "daf_truck_runway"
        model_name = "daf_xf_450_tractor"

        self._entities = {
            "joint_name" : [
                "front_left_steer_joint",
                "front_right_steer_joint",
                "rear_left_wheel_joint",
                "rear_right_wheel_joint",
                # "motor_1_joint",
                # "motor_2_joint",
                # "motor_3_joint",
                # "motor_4_joint",
                # "motor_5_joint",
                # "motor_6_joint",
                # "motor_f_joint",
                # "left_aileron_joint",
                # "right_aileron_joint",
                # "elevator_joint",
            ]
        }

        # Timer
        timer_period = 0.1  # seconds
        self.timer = self.create_timer(timer_period, self.timer_callback)

        # joint state publishers
        self._joint_state_publishers = {}
        for joint_name in self._entities["joint_name"]:
            self._joint_state_publishers[joint_name] = JointStatePublishers(
                  self, model_name, joint_name)

        # PID publishers
        self._pid_publishers = {}
        for joint_name in self._entities["joint_name"]:
            self._pid_publishers[joint_name] = PIDPublishers(
                self, model_name, joint_name)

        # Subscriptions (Ignition)
        self._ign_node = IgnNode()

        # subscription helper
        def subscribe(topic, callback, msg_type_name, opts):
          if self._ign_node.subscribe(topic, callback, msg_type_name, opts):
              self.get_logger().info(
                  "Subscribing to type [{}] on topic [{}]".format(
                      msg_type_name, topic))
          else:
              self.get_logger().error(
                  "Error subscribing to topic [{}]".format(topic))

        # subscribe to model topic
        self._last_model = Model()

        def sub_to_model(world_name, model_name):
          topic = "/world/{}/model/{}/joint_state".format(
              world_name, model_name)
          msg_type_name = Model.DESCRIPTOR.full_name
          opts = SubscribeOptions()

          # callback to capture the joint state msg
          def cb(msg: Model) -> None:
              self._last_model = msg

          subscribe(topic, cb, msg_type_name, opts)
        
        sub_to_model(world_name, model_name)

        self._pid_subscribers = {}
        for joint_name in self._entities["joint_name"]:
            self._pid_subscribers[joint_name] = PIDSubscribers(
                self, model_name, joint_name)

    def timer_callback(self):
        # Publish last model message
        model = self._last_model
        sim_time = model.header.stamp.sec + model.header.stamp.nsec/1.0E9

        stamp = builtin_interfaces.msg.Time(
            sec=model.header.stamp.sec,
            nanosec=model.header.stamp.nsec)
        # stamp = self.get_clock().now().to_msg()

        for joint in model.joint:
            if joint.name in self._entities["joint_name"]:
                joint_publisher = self._joint_state_publishers[joint.name]
                joint_publisher.pub_position(stamp, joint.axis1.position)
                joint_publisher.pub_velocity(stamp, joint.axis1.velocity)
                joint_publisher.pub_force(stamp, joint.axis1.force)
                joint_publisher.pub_acceleration(stamp, joint.axis1.acceleration)
                # print("joint_name:  {}".format(joint.name))
                # print("  type:      {}".format(joint.type))
                # print("  position:  {}".format(joint.axis1.position))
                # print("  velocity:  {}".format(joint.axis1.velocity))
                # print("  force:     {}".format(joint.axis1.force))
                # print("  accel:     {}".format(joint.axis1.acceleration))

        # Publish last PID message
        for joint in model.joint:
            if joint.name in self._entities["joint_name"]:
                sub = self._pid_subscribers[joint.name]
                pub = self._pid_publishers[joint.name]
                pid_act = sub.last_pid_act.data
                pid_tgt = sub.last_pid_tgt.data
                pid_cmd = sub.last_pid_cmd.data
                pub.pub_act(pid_act)
                pub.pub_tgt(pid_tgt)
                pub.pub_cmd(pid_cmd)
                # print("pid_act:     {}".format(pid_act))
                # print("pid_tgt:     {}".format(pid_tgt))
                # print("pid_cmd:     {}".format(pid_cmd))


def main(args=None):
    rclpy.init(args=args)

    model_bridge = ModelBridge()

    rclpy.spin(model_bridge)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    model_bridge.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()